from fnmatch import (
    fnmatch,
)

import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)

from labels.pkg.cataloger.elixir.parse_mix_exs import (
    parse_mix_exs,
)
from labels.pkg.cataloger.elixir.parse_mix_lock import (
    parse_mix_lock,
)
from labels.pkg.cataloger.generic.cataloger import (
    Request,
)


def on_next_elixir(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in (
                        "**/mix.lock",
                        "mix.lock",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_mix_lock,
                            parser_name="parse-elixir-mix-lock",
                        ),
                    )
                if any(
                    fnmatch(value, x)
                    for x in (
                        "**/mix.exs",
                        "mix.exs",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_mix_exs,
                            parser_name="parse-elixir-mix-exs",
                        ),
                    )

            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import logging
import os
import textwrap
from concurrent.futures import ThreadPoolExecutor
from typing import Unpack

import click

from labels.advisories.images.database import DATABASE
from labels.config.bugsnag import initialize_bugsnag
from labels.config.logger import configure_logger, modify_logger_level
from labels.core.options import MutuallyExclusiveOption, RequiredAnyCommand
from labels.core.sbom_configurator import ScanArgs, build_sbom_config
from labels.core.source_dispatcher import resolve_sbom_source
from labels.format import format_sbom
from labels.pkg.cataloger.complete import complete_package
from labels.pkg.operations.package_operation import package_operations_factory

LOGGER = logging.getLogger(__name__)


def show_banner() -> None:
    logo = textwrap.dedent(
        """
         ───── ⌝
        |    ⌝|  Fluid Attacks
        |  ⌝  |  We hack your software.
         ─────
        """,
    )
    click.secho(logo, fg="red")


@click.command(cls=RequiredAnyCommand, required_any=["o_from", "config"])
@click.argument("arg")
@click.option(
    "--from",
    "o_from",
    type=click.Choice(
        ["docker", "dir", "docker-daemon", "ecr"],
        case_sensitive=False,
    ),
    help=(
        "Source of the scan: 'docker' for scanning Docker images or 'dir' for scanning directories."
    ),
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(
        [
            "fluid-json",
            "cyclonedx-json",
            "spdx-json",
            "cyclonedx-xml",
            "spdx-xml",
        ],
        case_sensitive=False,
    ),
    default="fluid-json",
    help="Output format for the scanned data.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--output",
    "-o",
    help="Output filename.",
    default="sbom",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--docker-user",
    default=None,
    help="Docker registry username.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--docker-password",
    default=None,
    help="Docker registry password.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--aws-external-id",
    default=None,
    help="Docker registry username.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--aws-role",
    default=None,
    help="Docker registry password.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--config",
    "-c",
    is_flag=True,
    default=False,
    cls=MutuallyExclusiveOption,
    help="Path to an advanced configuration file with additional settings.",
    mutually_exclusive=[
        "aws-role",
        "aws-external-id",
        "docker-user",
        "docker-password",
        "output_format",
        "output",
        "o_from",
    ],
)
@click.option(
    "--debug",
    help="Run the application on debug mode",
    is_flag=True,
)
def scan(arg: str, **kwargs: Unpack[ScanArgs]) -> None:
    configure_logger()
    initialize_bugsnag()
    show_banner()

    sbom_config = build_sbom_config(arg, **kwargs)

    if sbom_config.debug or kwargs["debug"]:
        modify_logger_level()

    main_sbom_resolver = resolve_sbom_source(sbom_config)

    DATABASE.initialize()

    LOGGER.info("📦 Generating SBOM from %s: %s", sbom_config.source_type.value, sbom_config.source)
    packages, relationships = package_operations_factory(main_sbom_resolver)

    with ThreadPoolExecutor(
        max_workers=min(32, (os.cpu_count() or 1) * 5 if os.cpu_count() is not None else 32),
    ) as executor:
        LOGGER.info("📦 Gathering additional package information")
        packages = list(filter(None, executor.map(complete_package, packages)))

    LOGGER.info("📦 Preparing %s report", sbom_config.output_format)
    format_sbom(
        packages=packages,
        relationships=relationships,
        config=sbom_config,
        resolver=main_sbom_resolver,
    )


if __name__ == "__main__":
    scan(prog_name="labels")



import asyncio

from labels.internal.file_resolver.container_image import ContainerImage
from labels.model.core import SbomConfig, SourceType
from labels.sources.directory_source import Directory
from labels.sources.docker import get_docker_image, get_image_context
from labels.sources.ecr import AwsRole, ecr_connection
from labels.utils.exceptions import UnexpectedSBOMSourceError


def resolve_sbom_source(sbom_config: SbomConfig) -> Directory | ContainerImage:
    match sbom_config.source_type:
        case SourceType.DIRECTORY:
            return Directory(
                root=sbom_config.source,
                exclude=sbom_config.exclude,
            )
        case SourceType.DOCKER | SourceType.DOCKER_DAEMON:
            daemon = sbom_config.source_type == SourceType.DOCKER_DAEMON
            docker_image = get_docker_image(
                sbom_config.source,
                username=sbom_config.docker_user,
                password=sbom_config.docker_password,
                daemon=daemon,
            )
            if not docker_image:
                error_msg = f"No image found for {sbom_config.source}"
                raise ValueError(error_msg)

            context = get_image_context(
                image=docker_image,
                username=sbom_config.docker_user,
                password=sbom_config.docker_password,
                daemon=daemon,
            )
            if context is None:
                error_msg = f"No context found for {docker_image}"
                raise ValueError(error_msg)
            return ContainerImage(
                img=docker_image,
                context=context,
                lazy=False,
            )
        case SourceType.ECR:
            if not sbom_config.aws_role:
                error_msg = "The AWS role wasn't defined"
                raise ValueError(error_msg)
            role = AwsRole(
                external_id=sbom_config.aws_external_id,
                role=sbom_config.aws_role,
            )

            token, image_metadata = asyncio.run(
                ecr_connection(role, sbom_config.source),
            )

            if not image_metadata:
                error_msg = f"No image found for {sbom_config.source}"
                raise ValueError(error_msg)

            context = get_image_context(
                image=image_metadata,
                aws_creds=f"AWS:{token}",
            )
            if context is None:
                error_msg = f"No context found for {image_metadata}"
                raise ValueError(error_msg)
            return ContainerImage(
                img=image_metadata,
                context=context,
                lazy=False,
            )
        case _:
            raise UnexpectedSBOMSourceError



import logging
import re
from typing import Self

from packageurl import PackageURL
from univers.version_range import (
    RANGE_CLASS_BY_SCHEMES,
    InvalidVersionRange,
    VersionConstraint,
    VersionRange,
    split_req,
)
from univers.versions import (
    AlpineLinuxVersion,
    ComposerVersion,
    ConanVersion,
    DebianVersion,
    GolangVersion,
    InvalidVersion,
    MavenVersion,
    NugetVersion,
    PypiVersion,
    RpmVersion,
    RubygemsVersion,
    SemverVersion,
    Version,
)

from labels.model.core import (
    Advisory,
    Package,
)

LOGGER = logging.getLogger(__name__)


def _get_version_scheme_by_namespace(package: Package, namespace: str) -> Version | None:
    schemes = {
        "distro": {
            "alpine": AlpineLinuxVersion,
            "debian": DebianVersion,
            "redhat": RpmVersion,
            "ubuntu": DebianVersion,
        },
        "language": {
            "dart": SemverVersion,
            "dotnet": NugetVersion,
            "go": GolangVersion,
            "java": MavenVersion,
            "javascript": SemverVersion,
            "php": ComposerVersion,
            "python": PypiVersion,
            "ruby": RubygemsVersion,
            "rust": SemverVersion,
            "swift": SemverVersion,
        },
        "type": {
            "apk": AlpineLinuxVersion,
            "cocoapods": SemverVersion,
            "cargo": SemverVersion,
            "composer": ComposerVersion,
            "conan": ConanVersion,
            "deb": DebianVersion,
            "gem": RubygemsVersion,
            "golang": GolangVersion,
            "maven": MavenVersion,
            "npm": SemverVersion,
            "nuget": NugetVersion,
            "pub": SemverVersion,
            "pypi": PypiVersion,
            "rpm": RpmVersion,
            "swift": SemverVersion,
        },
    }

    def _get_language_scheme() -> Version | None:
        return schemes["language"].get(package.language.value)

    def _get_distro_scheme() -> Version | None:
        if package.p_url:
            package_url = PackageURL.from_string(package.p_url)
            if isinstance(package_url.qualifiers, dict) and (
                distro := package_url.qualifiers.get("distro_id")
            ):
                return schemes["distro"].get(distro)
        return None

    parts = namespace.split(":")
    if len(parts) < 3:
        return _get_language_scheme() or _get_distro_scheme()

    namespace_type, subtype = parts[1], parts[2]
    result = schemes.get(namespace_type, {}).get(subtype)

    return result or _get_language_scheme() or _get_distro_scheme()


class ApkVersionRange(VersionRange):  # type: ignore[misc]
    scheme = "apk"
    version_class = AlpineLinuxVersion

    @classmethod
    def from_native(cls, string: str) -> Self:
        constraints: list[str] = []
        match = re.match(r"([<>=~!^]*)(.*)", string)
        if not match:
            LOGGER.error("Invalid version range format: %s", string)
            return cls(constraints=constraints)
        comparator, version = match.groups()
        version = version.strip()
        return cls(
            constraints=[
                VersionConstraint(comparator=comparator, version=cls.version_class(version)),
            ],
        )


class PubVersionRange(VersionRange):  # type: ignore[misc]
    """Version range class for pub (not supported in univers yet).

    All restrictions and conditions based on:
    https://github.com/dart-lang/pub_semver/blob/master/README.md#semantics
    """

    scheme = "pub"
    version_class = SemverVersion

    vers_by_native_comparators = {  # noqa: RUF012
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
        "=": "=",
    }

    @classmethod
    def from_native(cls, string: str) -> Self:
        constraints = []
        comparator = ""

        for constraint_item in string.split():
            if not re.match(
                r"^[<>=~!^]*\d+(\.\d+)*(-[a-zA-Z0-9]+)?(\+[a-zA-Z0-9]+)?$",
                constraint_item,
            ):
                continue

            if constraint_item.startswith("^"):
                base_version = cls.version_class(constraint_item.lstrip("^"))
                if base_version.major > 0:
                    upper = cls.version_class(f"{base_version.major + 1}.0.0")
                else:
                    upper = cls.version_class(f"0.{base_version.minor + 1}.0")

                upper = cls.version_class(str(upper).split("-")[0].split("+")[0])
                lower = base_version
                constraints.extend(
                    [
                        VersionConstraint(comparator=">=", version=lower),
                        VersionConstraint(comparator="<", version=upper),
                    ],
                )
                continue
            comparator, version = VersionConstraint.split(constraint_item)
            constraints.append(
                VersionConstraint(comparator=comparator, version=cls.version_class(version)),
            )
            comparator = ""

        return cls(constraints=constraints)


class ComposerVersionRange(VersionRange):  # type: ignore[misc]
    """Version range class for pub (not supported in univers yet).

    All restrictions and conditions based on:
    https://github.com/dart-lang/pub_semver/blob/master/README.md#semantics
    """

    scheme = "composer"
    version_class = ComposerVersion

    vers_by_native_comparators = {  # noqa: RUF012
        "==": "=",
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
        "=": "=",
    }

    @classmethod
    def build_constraint_from_string(cls, string: str) -> Self:
        comparator, version = split_req(
            string=string,
            comparators=cls.vers_by_native_comparators,
            strip=")(",
        )
        version = cls.version_class(version)
        return VersionConstraint(comparator=comparator, version=version)

    @classmethod
    def from_native(cls, string: str) -> Self:
        return cls(constraints=[cls.build_constraint_from_string(string)])


def convert_to_maven_range(constraint: str) -> str:
    """Convert a version constraint to Maven-compatible range format.

    According to the rules:
    https://maven.apache.org/enforcer/enforcer-rules/versionRanges.html
    """
    constraint = constraint.strip()
    result = ""
    match constraint[:2]:
        case "<=":
            version = constraint[2:].strip()
            result = f"(,{version}]"
        case ">=":
            version = constraint[2:].strip()
            result = f"[{version},)"
        case _:
            match constraint[:1]:
                case "<":
                    version = constraint[1:].strip()
                    result = f"(,{version})"
                case ">":
                    version = constraint[1:].strip()
                    result = f"({version},)"
                case "=":
                    version = constraint[1:].strip()
                    result = f"[{version}]"
                case _:
                    result = f"[{constraint},)"
    return result


def normalize_npm_constraint(constraint: str) -> str:
    constraint = re.sub(r"(\d)\.x-(\d)\.x", r"\1.x - \2.x", constraint)
    return constraint.strip()


def _normalize_version_constraint(constraint: str, scheme: str) -> str:
    """Normalize a version constraint based on the specified versioning scheme.

    This function normalizes version constraints for various schemes. Currently,
    the following schemes are supported:
    - npm: Fixes invalid NPM range formats (e.g., '5.x-1.x' to '5.x - 1.x').
    - maven, nuget: Converts constraints to a Maven-compatible range syntax.

    Additional schemes can be added as needed.

    Args:
        constraint (str): The raw version constraint string.
        scheme (str): The versioning scheme (e.g., 'npm', 'maven').

    Returns:
        str: The normalized version constraint.

    """
    normalizers = {
        "npm": normalize_npm_constraint,
        "maven": convert_to_maven_range,
        "nuget": convert_to_maven_range,
    }
    normalizer = normalizers.get(scheme)
    if normalizer:
        return normalizer(constraint)
    return constraint


def _compare_single_constraint(version: Version, constraint: str, scheme: str) -> bool:
    version_range: VersionRange | None = {
        **RANGE_CLASS_BY_SCHEMES,
        "apk": ApkVersionRange,
        "pub": PubVersionRange,
        "composer": ComposerVersionRange,
    }.get(scheme)

    if not version_range:
        LOGGER.error(
            "Invalid version scheme: %s",
            scheme,
        )
        return False
    try:
        constraint = _normalize_version_constraint(constraint.strip(), scheme)
        return version in version_range.from_native(constraint)
    except (InvalidVersion, InvalidVersionRange, TypeError):
        return False


def _matches_constraint(version: Version, constraint: str, version_scheme: str) -> bool:
    if not constraint:
        return True

    constraints = constraint.split(",")
    return all(
        _compare_single_constraint(version, constraint.strip(), version_scheme)
        for constraint in constraints
    )


def matches_version(package: Package, advisory: Advisory) -> bool:
    version_type = _get_version_scheme_by_namespace(package, advisory.namespace)
    if version_type is None:
        LOGGER.debug(
            "No version scheme found for namespace %s",
            advisory.namespace,
        )
        return False

    if advisory.version_constraint is None:
        return True
    if not package.p_url:
        return False

    try:
        match = re.match(r"([<>=~!^]*)(.*)", package.version)
        if not match:
            return False

        _, version = match.groups()
        version = version.strip()

        return any(
            _matches_constraint(
                version_type(version),
                constraint.strip(),
                PackageURL.from_string(package.p_url).type,
            )
            for constraint in advisory.version_constraint.split("||")
        )
    except (AttributeError, InvalidVersion):
        return False



from collections.abc import Mapping
from typing import Any

from click import Command, Context, Option, UsageError


class RequiredAnyCommand(Command):
    def __init__(self, *args: Any, required_any: list[str], **kwargs: Any) -> None:
        self.required_any = required_any or []
        super().__init__(*args, **kwargs)

    def invoke(self, ctx: Context) -> Any:
        if self.required_any and not any(ctx.params.get(opt) for opt in self.required_any):
            required_opts = ", ".join(
                next(
                    (p.opts[0] for p in ctx.command.params if p.name == opt),
                    opt,
                )
                for opt in self.required_any
            )
            error_msg = f"You must provide at least one of the following options: {required_opts}."
            raise UsageError(error_msg)
        return super().invoke(ctx)


class MutuallyExclusiveOption(Option):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help_msg = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = help_msg + (
                f" NOTE: This argument is mutually exclusive with arguments: [{ex_str}]."
            )
        super().__init__(*args, **kwargs)

    def handle_parse_result(
        self,
        ctx: Context,
        opts: Mapping[str, Any],
        args: list[str],
    ) -> tuple[Any, list[str]]:
        command_name = self.opts[0] if self.opts else self.name
        mutually_exclusive_names = {
            next((p.opts[0] for p in ctx.command.params if p.name == opt), opt)
            for opt in self.mutually_exclusive
        }
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            error_msg = (
                f"Illegal usage: `{command_name}` is mutually exclusive with arguments "
                f"`{', '.join(mutually_exclusive_names)}`."
            )
            raise UsageError(error_msg)

        return super().handle_parse_result(ctx, opts, args)
